/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019,2021,2023 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_nls.h"
#include "cff_charconv.h"

/*
#include <gtk/gtk.h>
#include "cff_tools-gtk.h"
*/


enum
{
	kColumnMenu,
	kColumnKey,
	kColumnRowData,
	kNumberOfColumns
};


/* cff_gui_adjust_menu => cff_gui.c */

#ifdef HAVE_GMENU
extern gboolean cff_gui_adjust_menu(GSimpleAction *action,
					GVariant *p, gpointer data);
#define AccelDoAdjustGuiMenu(action) \
		cff_gui_adjust_menu((action), NULL, NULL)
#else
extern gboolean cff_gui_adjust_menu(GtkWidget *action, gpointer data);
#define AccelDoAdjustGuiMenu(action) \
		cff_gui_adjust_menu((action), NULL)
#endif

#ifdef HAVE_GMENU
#define CffAccelGroup	GtkApplication
#else
#define CffAccelGroup	GtkAccelGroup
#endif

#ifdef HAVE_GMENU
#define AccelDoAddAction(accel, group) \
	cff_accel_set_for_action((accel), (group), TRUE)
#define AccelDoRemoveAction(accel, group) \
	cff_accel_set_for_action((accel), (group), FALSE)
#else
#define AccelDoAddAction(accel, group) \
	cff_accel_add_accelerator((accel), (group))
#define AccelDoRemoveAction(accel, group) \
	cff_accel_remove_accelerator((accel), (group))
#endif


static CffAccelGroup *gAccelGroup = NULL;

static cff_accel_t gAccelTab[kAccelKeyTotal] =
{
	{ "FileOpen",		NULL, NULL, 0, 0, 0 },
	{ "FileSave",		NULL, NULL, 0, 0, 0 },
	{ "FileImage",		NULL, NULL, 0, 0, 0 },
	{ "FileClose",		NULL, NULL, 0, 0, 0 },
	{ "FileInfo",		NULL, NULL, 0, 0, 0 },
	{ "FileSaverc",		NULL, NULL, 0, 0, 0 },
	{ "FileQuit",		NULL, NULL, 0, 0, 0 },
	{ "EditUndo",		NULL, NULL, 0, 0, 0 },
	{ "EditRedo",		NULL, NULL, 0, 0, 0 },
	{ "EditLoad",		NULL, NULL, 0, 0, 0 },
	{ "EditSave",		NULL, NULL, 0, 0, 0 },
	{ "EditClear",		NULL, NULL, 0, 0, 0 },
	{ "EditAdd",		NULL, NULL, 0, 0, 0 },
	{ "EditRemove",		NULL, NULL, 0, 0, 0 },
	{ "OptAccel",		NULL, NULL, 0, 0, 0 },
	{ "OptKeyframe",	NULL, NULL, 0, 0, 0 },
	{ "OptIframe",		NULL, NULL, 0, 0, 0 },
	{ "OptPframe",		NULL, NULL, 0, 0, 0 },
	{ "OptCursor",		NULL, NULL, 0, 0, 0 },
	{ "OptOverwrite",	NULL, NULL, 0, 0, 0 },
	{ "OptDelete",		NULL, NULL, 0, 0, 0 },
	{ "OptBeep",		NULL, NULL, 0, 0, 0 },
	{ "OptSave",		NULL, NULL, 0, 0, 0 },
	{ "HelpFFmpeg",		NULL, NULL, 0, 0, 0 },
	{ "HelpCutterff",	NULL, NULL, 0, 0, 0 },
	{ "HelpAbout",		NULL, NULL, 0, 0, 0 }

}; /* static cff_accel_t gAccelTab[kAccelKeyTotal] */

typedef struct
{
	cff_accel_t	*rowdata;
	cff_accel_t	accel_tab[kAccelKeyTotal];
	cff_ListStore_t	list;
	GtkTreeIter	iter;

} CffAccelAssignDialog;


static cff_accel_t *cff_accel_find_by_name(cff_accel_t *accelp, const gchar *name)
{
	gint loop = kAccelKeyTotal - 1;

	do
	{
		accelp--;

		if (strcmp(accelp->name, name) == 0)
		{
			return accelp;
		}
	}
	while (--loop > kAccelKeyNone);

	return NULL;

} /* static cff_accel_t *cff_accel_find_by_name(cff_accel_t*, const gchar*) */

static cff_accel_t *cff_accel_find_by_keys(cff_accel_t *accelp, guint keyval,
						GdkModifierType modifier)
{
	gint loop = kAccelKeyTotal - 1;

	do
	{
		accelp--;

		if (keyval == accelp->keyval &&
				modifier == accelp->modifier)
		{
			return accelp;
		}
	}
	while (--loop > kAccelKeyNone);

	return NULL;

} /* static cff_accel_t *cff_gaccel_find_by_keys(cff_accel_t*, guint, GdkModifierType) */


#ifdef HAVE_GMENU
static gboolean cff_accel_key_pressed(GtkEventControllerKey *cntrl,
			guint keyval, guint keycode, GdkModifierType state,
							cff_accel_t *accelp)
#else
static gboolean cff_accel_key_pressed(GtkWidget *window,
				GdkEventKey *event, cff_accel_t *accelp)
#endif
{

#ifndef HAVE_GMENU
	guint state;

	if (event->length == 0)
	{
		return FALSE;
	}

	state = event->state;
#endif
	if (state & GDK_CONTROL_MASK)
	{

#ifndef HAVE_GMENU
		guint keyval = event->keyval;
#endif
		accelp = cff_accel_find_by_keys(accelp, keyval, state);

		if (accelp)
		{
			/* TRUE: busy => disabled */
			return AccelDoAdjustGuiMenu(accelp->action);
		}

	} /* if (state & GDK_CONTROL_MASK) */

	return FALSE;

} /* static gboolean cff_accel_key_pressed(GtkEventControllerKey*, guint,
					guint, GdkModifierType, cff_accel_t*) */

#ifdef HAVE_GMENU

static void cff_accel_set_for_action(const cff_accel_t *accelp,
				CffAccelGroup *accel_group, gboolean add)
{
	/*
		To remove all accelerators for an action, use
		an empty, zero-terminated array for accels
		GTK-3 => NULL -- GTK-4 => ""
	*/

	const gchar	*action_name;
	gchar		*accel[2], app_name[256];

	action_name = g_action_get_name(G_ACTION(accelp->action));

	g_snprintf(app_name, sizeof(app_name), "%s%s",
			CFF_GACTION_PREFIX_STRING, action_name);

	if (add == FALSE)
	{

#if GTK_MAJOR_VERSION == 3
		accel[0] = NULL;
#else
		accel[0] = "";
#endif
	}

	else
	{
		accel[0] = gtk_accelerator_name(accelp->key, accelp->modifier);

	} /* if (add == FALSE) .. else */

	accel[1] = NULL;

	gtk_application_set_accels_for_action(accel_group,
					app_name, (const gchar **) accel);
	if (add)
	{
		g_free(accel[0]);
	}

} /* static void cff_accel_set_for_action(const cff_accel_t*,
						CffAccelGroup*, gboolean) */
#else /* HAVE_GMENU */

static void cff_accel_remove_accelerator(const cff_accel_t *accelp,
						CffAccelGroup *accel_group)
{
	gtk_accel_group_unlock(accel_group);

	gtk_widget_remove_accelerator(accelp->action, accel_group,
						accelp->key, accelp->modifier);
	gtk_accel_group_lock(accel_group);

} /* static void cff_accel_remove_accelerator(const cff_accel_t*, CffAccelGroup*) */

static void cff_accel_add_accelerator(const cff_accel_t *accelp,
						CffAccelGroup *accel_group)
{
	gtk_accel_group_unlock(accel_group);

	gtk_widget_add_accelerator(accelp->action, "activate", accel_group,
				accelp->key, accelp->modifier,  GTK_ACCEL_VISIBLE);

	gtk_accel_group_lock(accel_group);

} /* static void cff_accel_add_accelerator(const cff_accel_t*, CffAccelGroup*) */

#endif /* #ifdef HAVE_GMENU .. #else */


void cff_accel_bind_widget(gint accel_id, const gchar *label, GuiMenuAction *action, GtkWidget *window)
{
	CffAccelGroup	*accel_group;
	cff_accel_t	*accelp;

	if (accel_id == kAccelKeyNone || action == NULL)
	{
		return;
	}

	accel_group = gAccelGroup;

	if (accel_group == NULL)
	{

#ifdef HAVE_GMENU

		GtkEventController *cntrl_key;

#if GTK_MAJOR_VERSION == 3
		cntrl_key = gtk_event_controller_key_new(window);
#else
		cntrl_key = gtk_event_controller_key_new();
		gtk_widget_add_controller(window, cntrl_key);
#endif
		/*
			GTK_PHASE_CAPTURE => runs from the
			toplevel down to the event widget
		*/

		gtk_event_controller_set_propagation_phase(cntrl_key,
							GTK_PHASE_CAPTURE);

		GuiSignalConnect(cntrl_key, "key-pressed",
			cff_accel_key_pressed, &gAccelTab[kAccelKeyTotal]);

		accel_group = gtk_window_get_application(GTK_WINDOW(window));

#else /* ! HAVE_GMENU */

		accel_group = gtk_accel_group_new();
		gtk_accel_group_lock(accel_group);

		gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);

		GuiSignalConnect(window, "key-press-event",
			cff_accel_key_pressed, &gAccelTab[kAccelKeyTotal]);

#endif /* #ifdef HAVE_GMENU .. #else */

		gAccelGroup = accel_group;
	}

	accelp = &gAccelTab[accel_id];

	accelp->label	= label;
	accelp->action	= action;

	if (accelp->key)
	{
		AccelDoAddAction(accelp, accel_group);
	}

} /* void cff_gaccel_bind_widget(gint, const gchar*, GtkWidget*, GtkWidget*) */

static void cff_accel_dialog_okay(CffAccelAssignDialog *acceldlg)
{
	gint	 	loop;
	cff_accel_t	*old, *mod;

	old	= gAccelTab;
	mod 	= acceldlg->accel_tab;
	loop	= 0;
	
	do
	{
		if (old->keyval != mod->keyval || old->modifier != mod->modifier)
		{
			if (old->keyval)
			{
				AccelDoRemoveAction(old, gAccelGroup);
			}
		}

		old++;
		mod++;
	}
	while (++loop < kAccelKeyTotal);

	old	= gAccelTab;
	mod 	= acceldlg->accel_tab;
	loop	= 0;
	
	do
	{
		if (old->keyval != mod->keyval || old->modifier != mod->modifier)
		{
			if (mod->keyval)
			{
				AccelDoAddAction(mod, gAccelGroup);
			}

			old->keyval   = mod->keyval;
			old->key      = mod->key;
			old->modifier = mod->modifier;
		}

		old++;
		mod++;
	}
	while (++loop < kAccelKeyTotal);

} /* static void cff_accel_dialog_okay(GtkWidget*, CffAccelAssignDialog*) */

#ifdef HAVE_GMENU
static gboolean cff_accel_dialog_key(GtkEventControllerKey *cntrl,
			guint keyval, guint keycode, GdkModifierType state,
						CffAccelAssignDialog *acceldlg)
#else
static gboolean cff_accel_dialog_key(GtkWidget *window,
			GdkEventKey *event, CffAccelAssignDialog *acceldlg)
#endif
{
	cff_accel_t *accelp;

#ifndef HAVE_GMENU
	guint keyval, state;
	keyval = event->keyval;
#endif

	if (keyval == Gui_KEY_BackSpace)
	{
		accelp = acceldlg->rowdata;

		if (accelp->keyval)
		{
			gtk_list_store_set(acceldlg->list.store,
				&(acceldlg->iter), kColumnKey, NULL, -1);

			accelp->keyval   = 0;
			accelp->key      = 0;
			accelp->modifier = 0;
		}

		return TRUE;
	}

#ifdef HAVE_GMENU
	else if (state & GDK_CONTROL_MASK)
#else
	else if (event->length && ((state = event->state) & GDK_CONTROL_MASK))
#endif
	{
		if (cff_accel_find_by_keys(&acceldlg->accel_tab[kAccelKeyTotal],
								keyval, state))
		{
			cff_guitl_beep();
		}

		else
		{
			gint key = keyval;

			if (state & GDK_SHIFT_MASK)
			{
				key = gdk_keyval_to_lower(key);
			}

			if (gtk_accelerator_valid(key, state))
			{
				gchar *name = gtk_accelerator_name(key, state);

				if (name != NULL)
				{
					gchar *str;

					str = cff_charconv_utf8_from_locale(name, NULL);

					gtk_list_store_set(acceldlg->list.store,
						&(acceldlg->iter), kColumnKey, str, -1);
					CFF_CHARCONV_UTF8_FREE(str, name);
					g_free(name);

					accelp = acceldlg->rowdata;

					accelp->keyval   = keyval;
					accelp->key      = key;
					accelp->modifier = state;

				} /* if (name != NULL) */

			} /* if (gtk_accelerator_valid(key, state)) */

		} /* if (cff_gaccel_find_by_keys(..) .. else */

		return TRUE;

	} /* if (keyval == Gui_KEY_BackSpace) .. else if (..) */

	return FALSE;

} /* static gboolean cff_accel_dialog_key(GtkWidget*, GdkEventKey*, CffAccelAssignDialog*) */

static void cff_accel_dialog_row_selected(GtkTreeSelection *selection, CffAccelAssignDialog *acceldlg)
{
	GtkTreeIter	iter;
	GtkTreeModel	*model;

	if (gtk_tree_selection_get_selected(selection, &model, &iter))
	{
		acceldlg->iter = iter;
		gtk_tree_model_get(model, &iter, kColumnRowData, &(acceldlg->rowdata), -1);
	}

} /* static void cff_accel_dialog_row_selected(GtkTreeSelection*, CffAccelAssignDialog*) */

static void cff_gaccel_build_column(cff_ListStore_t *list, gint colnr, const gchar *title)
{
	gchar			*str;
	GtkTreeViewColumn	*column;
	GtkCellRenderer		*renderer;

	column = gtk_tree_view_column_new();

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", colnr, NULL);

	gtk_tree_view_append_column(list->view, column);

	str = cff_charconv_utf8_gettext(title);
	gtk_tree_view_column_set_title(column, str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);

} /* static void cff_gaccel_build_column(cff_ListStore_t*, gint, const gchar*) */

void cff_accel_dialog_run(void)
{
	gint			loop;
	GtkResponseType		response;
	GtkWidget		*frame, *label, *swin, *vbox, *dialog;
	CffAccelAssignDialog	acceldlg;
#ifdef HAVE_GMENU
	GtkEventController	*controller;
#endif

	dialog = cff_guitl_dialog_new_cancel_ok(gettext_nop("CutterFF: Keyboard"),
								gWindata.mainwin);
	vbox = GuiVBoxNew(FALSE, 2);

	acceldlg.list.store =
		gtk_list_store_new(kNumberOfColumns, G_TYPE_STRING,
					G_TYPE_STRING, G_TYPE_POINTER);

	acceldlg.list.view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(acceldlg.list.store));
	g_object_unref(G_OBJECT(acceldlg.list.store));

	acceldlg.list.selection = gtk_tree_view_get_selection(acceldlg.list.view);

	gtk_tree_selection_set_mode(acceldlg.list.selection,
						GTK_SELECTION_BROWSE);

	GuiSignalConnect(acceldlg.list.selection, "changed",
				cff_accel_dialog_row_selected, &acceldlg);

	swin = cff_guitl_scrolled_window_new(320, 180,
					GTK_WIDGET(acceldlg.list.view));
	GuiBoxAddChild(vbox, swin);

	cff_gaccel_build_column(&(acceldlg.list), kColumnMenu, gettext_nop("Menu"));
	cff_gaccel_build_column(&(acceldlg.list), kColumnKey, gettext_nop("Key"));

	memcpy(acceldlg.accel_tab, gAccelTab, sizeof(gAccelTab));

	loop = 0;

	do
	{
		gchar		*str[2], *title;
		GtkTreeIter	iter;

		str[0] = cff_charconv_utf8_gettext(acceldlg.accel_tab[loop].label);
		title = gtk_accelerator_name(acceldlg.accel_tab[loop].key,
						acceldlg.accel_tab[loop].modifier);
		str[1] = cff_charconv_utf8_from_locale(title, NULL);

		gtk_list_store_append(acceldlg.list.store, &iter);
		gtk_list_store_set(acceldlg.list.store, &iter,
			kColumnMenu, str[0],
			kColumnKey, str[1],
			kColumnRowData, &(acceldlg.accel_tab[loop]),
			-1);

		CFF_CHARCONV_UTF8_FREE(str[0], acceldlg.accel_tab[loop].label);
		CFF_CHARCONV_UTF8_FREE(str[1], title);
		g_free(title);

		if (loop == 0)
		{
			gtk_tree_selection_select_iter(acceldlg.list.selection, &iter);
		}
	}
	while (++loop < kAccelKeyTotal);

	g_object_set(acceldlg.list.view,
			"headers-visible", TRUE,
			"headers-clickable", FALSE,
			NULL);

	label = cff_guitl_label_new_gettext(
			gettext_nop("Select a menu entry and press your prefered keys,"
							" e.g. Quit => 'Cntrl' + 'Q'"));
	GuiLabelSetLineWrap(label, TRUE);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, label);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 1);

#ifdef HAVE_GMENU

#if GTK_MAJOR_VERSION == 3
	controller = gtk_event_controller_key_new(dialog);
#else
	gtk_widget_show(dialog);
	controller = gtk_event_controller_key_new();
	gtk_widget_add_controller(dialog, controller);
#endif
	GuiSignalConnect(controller, "key-pressed",
					cff_accel_dialog_key, &acceldlg);

#else /* ! HAVE_GMENU */

	GuiSignalConnect(dialog, "key-press-event",
					cff_accel_dialog_key, &acceldlg);

#endif /* #ifdef HAVE_GMENU .. #else */

	cff_guitl_dialog_set_child(dialog, vbox);
	GuiWidgetShowAll(vbox);

	response = GuiDialogRun(dialog);

	if (response == GTK_RESPONSE_ACCEPT)
	{
		cff_accel_dialog_okay(&acceldlg);
	}

	GuiDialogClose(dialog, response);

} /* void cff_accel_dialog_run(void) */

gint cff_accel_parse_str(gchar *str)
{
	gchar *ptr = strchr(str, '=');

	if (ptr != NULL)
	{
		gchar *peq, *name;

		peq = ptr,
		name = str;
		*(ptr++) = '\0';

		if (*(ptr++) == '"')
		{
			gchar *accelerator = ptr;

			ptr = strrchr(accelerator, '"');

			if (ptr != NULL)
			{
				guint		accel_key;
				GdkModifierType	modifier;

				*ptr      = '\0';
				modifier  = 0;
				accel_key = 0;

				gtk_accelerator_parse(accelerator, &accel_key, &modifier);

				if (accel_key != 0 && (modifier & GDK_CONTROL_MASK))
				{
					cff_accel_t *accelp;

					accelp = cff_accel_find_by_name(&gAccelTab[kAccelKeyTotal], name);

					if (accelp != NULL)
					{
						accelp->modifier = modifier;
						accelp->key = accel_key;

						accelp->keyval = (modifier & GDK_SHIFT_MASK) ?
						gdk_keyval_to_upper(accel_key) : accel_key;

						return TRUE;
					}

				} /* if (accel_key != 0 && (modifier & GDK_CONTROL_MASK)) */

				*ptr = '"';

			} /* if (ptr != NULL) */

		} /* if (*(ptr++) != '"') */

		*peq = '=';

	} /* if (ptr != NULL) */

	return FALSE;

} /* gint cff_accel_parse_str(gchar*) */

void cff_accel_save_data(FILE *fd, const gchar *id)
{
	cff_accel_t *accelp = gAccelTab;

	gint loop = kAccelKeyTotal - 1;

	do
	{
		if (accelp->key)
		{
			gchar *name = gtk_accelerator_name(accelp->key,
							accelp->modifier);
			if (name != NULL)
			{
				fprintf(fd, "%s %s=\"%s\"\n", id,
							accelp->name, name);
				g_free(name);
			}
		}

		accelp++;
	}
	while (--loop > kAccelKeyNone);

} /* void cff_accel_save_data(FILE*, const gchar*) */

/* EOF */
