/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017..2019,2021,2023,2024 Harald Foerster */
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#ifndef CFF_TOOLS_GTK__H
#define CFF_TOOLS_GTK__H

#include <gtk/gtk.h>


/* Works also with GTK-3 (newer versions) */
#if GTK_MAJOR_VERSION == 4
#define HAVE_GMENU	1
#endif

#ifdef HAVE_GMENU
#define GuiMenu			GMenu
#define GuiMenuItem		GMenuItem
#define GuiMenuModel		GMenuModel
#define GuiMenuAction		GSimpleAction
#else
#define GuiMenu			GtkWidget
#define GuiMenuItem		GtkWidget
#define GuiMenuModel		GtkWidget
#define GuiMenuAction		GtkWidget
#endif

#ifdef HAVE_GMENU
#define CFF_GACTION_GROUP_STRING	"app"
#define CFF_GACTION_PREFIX_STRING	CFF_GACTION_GROUP_STRING"."
#define CFF_GACTION_PREFIX_LENGTH	4
#endif


#define GUI_LIST_ROW_HEIGHT_MIN	16

typedef struct
{
	GtkListStore		*store;
	GtkTreeView		*view;
	GtkTreeSelection	*selection;

} cff_ListStore_t;

typedef struct
{
	GtkWidget	*topwin;  /* Window with focus */
	GtkWidget	*mainwin; /* This (cff_gui-gtk.c) window */
	GSList		*slist;   /* List of opened windows */

} cff_windata_t;

extern cff_windata_t gWindata;

#ifdef HAVE_GMENU
extern GMainLoop *gMainLoop;
#endif


/*
	------------
	 Prototypes
	------------
*/

GtkWidget *cff_guitl_window_new(GtkWidget *child, gint border);
GtkWidget *cff_guitl_window_new_gettext(const gchar *title, GtkWidget *child, gint border);
void       cff_guitl_window_set_title_filename(GtkWidget *window, const gchar *title);
void       cff_guitl_window_setup(GtkWidget *widget, gint prop);
void       cff_guitl_window_set_topwin(GtkWidget *window);
GtkWidget *cff_guitl_scrolled_window_new(gint wmin, gint hmin, GtkWidget *child);

void cff_guitl_dialog_set_child(GtkWidget *dialog, GtkWidget *child);
void cff_guitl_dialog_set_modal(GtkWidget *dialog, GtkWidget *parent, GtkWidget **mbar);

void     cff_guitl_dialog_info(const gchar *text, GtkWidget **mbar);
void     cff_guitl_dialog_error(const gchar *text, GtkWidget **mbar);
void     cff_guitl_dialog_warning(const gchar *text, GtkWidget **mbar);
gboolean cff_guitl_dialog_warning_2btn(const gchar *text, GtkWidget *parent, GtkWidget **mbar);
void     cff_guitl_dialog_okay_run(const gchar *title, GtkWidget *child, GtkWidget **mbar, int prop);

GtkWidget *cff_guitl_dialog_new_gettext_2btn(const gchar *title, const gchar *btn_no, const gchar *btn_yes, GtkWidget *parent);
GtkWidget *cff_guitl_dialog_new_cancel_ok(const gchar *title, GtkWidget *parent);
GtkWidget *cff_guitl_label_new_gettext(const gchar *str);
GuiMenuItem *cff_guitl_menu_item_new_gettext(const gchar *label);
GuiMenuItem *cff_guitl_check_menu_item_new_gettext(const gchar *label);
GtkWidget *cff_guitl_frame_new_add_child_gettext(const gchar *title, gint border, GtkWidget *child);
GtkWidget *cff_guitl_check_button_new_gettext(const gchar *label);
GtkWidget *cff_guitl_vbox_add_check_button_gettext(GtkWidget *vbox, const gchar *label);
GtkWidget *cff_guitl_hbox_add_button(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data);
GtkWidget *cff_guitl_vbox_add_button(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data);
GtkWidget *cff_guitl_hbox_add_button_gettext(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data);
GtkWidget *cff_guitl_vbox_add_button_gettext(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data);
GtkWidget *cff_guitl_hbutton_box_new(GtkWidget *parent, gint spacing);
GtkWidget *cff_guitl_button_box_add_button(GtkWidget *button_box, const gchar *text, GCallback func, gpointer data);
void       cff_guitl_beep(void);
void       cff_guitl_widget_set_tooltip_gettext(GtkWidget *widget, const gchar *text);
void       cff_guitl_widget_size_to_prop(GtkWidget *widget, gint prop);

#if GTK_MAJOR_VERSION > 2
GtkWidget *cff_guitl_box_new(GtkOrientation orientation,
					gboolean homogeneous, gint spacing);
#endif

#if GTK_MAJOR_VERSION == 4
void  cff_guitl_widget_set_border_width(GtkWidget *widget, gint width);
gint  cff_guitl_dialog_run(GtkWidget *dialog);
void  cff_guitl_main_level_start(void);
void  cff_guitl_main_level_quit(void);
#endif

#ifdef HAVE_GMENU
char *cff_guitl_create_action_name(gchar *dst, gsize len, const gchar *src);
#endif

/* Misc */
gboolean    cff_tl_is_ascii_str(const gchar *str);
gchar       *cff_tl_replace_non_utf8_chars_in_str(gchar *str);
void         cff_tl_file_path_local_to_unix(gchar *path);
void         cff_tl_file_path_unix_to_local(gchar *path);
const gchar *cff_tl_file_get_basename(const gchar *path);
GFile       *cff_tl_file_get_directory(const gchar *fname);
gboolean     cff_tl_file_is_directory(GFile *file);
gboolean     cff_tl_file_is_equal(GFile *file_src, GFile *file_dst);
#if GTK_MAJOR_VERSION == 4
GFile       *cff_tl_file_append_fname(GFile *dir, const gchar *fname);
#endif

#endif /* CFF_TOOLS_GTK__H */
